import {
  Class,
  Method,
} from '@overflow/commons/core/reflect';
import {
  ClassType, PropertyKeyType, TypeUtil,
} from '@overflow/commons/core/type';
import { ApplicationContext } from './context';
import {
  ConfigurationAnnotation,
  InstanceAnnotation,
  RunnerAnnotation,
  WebApplicationAnnotation,
} from './decorators';

export class Application {
  private _appClass: Class;
  private _appAnnotation: WebApplicationAnnotation;
  private _applicationContext: ApplicationContext;

  public constructor(primaryClass: ClassType) {
    this.initialize(primaryClass);
  }

  private initialize(primaryClass: ClassType): void {
    let clazz: Class = Class.forClassType(primaryClass);

    let wa: WebApplicationAnnotation = clazz.getOwnAnnotation(WebApplicationAnnotation);
    if (undefined === wa) {
      throw new Error(`Class is not WebApplication type. add @WebApplication annotation to class[${primaryClass.name}]`);
    }

    this._appClass = clazz;
    this._appAnnotation = wa;

    this._applicationContext = new ApplicationContext();

    let injectables: ClassType[] = this._appAnnotation.attributes.injectables;
    this._applicationContext.registerInjectables(...injectables);

    let configurations: ClassType[] = this._appAnnotation.attributes.configurations;
    this._applicationContext.registerConfigurations(...configurations);

    let jsonSources: any[] = this._appAnnotation.attributes.jsonSources;
    this._applicationContext.registerJSONSources(...jsonSources);

    this._applicationContext.registerInjectable(this._appClass.getType(), undefined);
  }

  public static async run(primaryClass: ClassType): Promise<ApplicationContext> {
    return new Application(primaryClass).run();
  }

  /**
   * run
   */
  public run(): ApplicationContext {

    let runner = this.findRunner();
    if (undefined === runner) {
      throw new Error(`There is not exist @Runner on Application[${this._appClass.getName()}]`);
    }

    let instance = this._applicationContext.getInstance(undefined, this._appClass.getType());

    runner.invoke(instance);

    return this._applicationContext;
  }

  private getInjectables(): ClassType[] {
    let injectables: Set<ClassType> = new Set();
    injectables.add(this._appClass.getType());
    this._appAnnotation.attributes.injectables.forEach(injectableType => {
      injectables.add(injectableType);
    });
    this._appAnnotation.attributes.configurations.forEach(configurationType => {
      injectables.add(configurationType);
    });

    return Array.from(injectables);
  }

  private findRunner(): Method | undefined {
    let methods = this._appClass.getOwnMethods();

    for (let key of Array.from(methods.keys())) {
      let method = methods.get(key);
      let runnerAnnotation = method.getOwnAnnotation(RunnerAnnotation);
      if (undefined === runnerAnnotation) {
        continue;
      }
      return method;
    }

    return undefined;
  }

//   private createContext(): ApplicationContext {
//     let clazz: Class = Class.forClassType(this._primaryClass);


//     let wa: WebApplicationAnnotation = clazz.getOwnAnnotation(WebApplicationAnnotation);
//     if (undefined === wa) {
//       throw new Error(`Class is not WebApplication type. add @WebApplication annotation to class[${this._primaryClass.name}]`);
//     }
//     this._applicationContext = new ApplicationContext(wa.attributes.injectables);

//     this.registerJSONSources(wa.attributes.jsonSources);
//     this.registerConfigurations(wa.attributes.configurations);

//     return this._applicationContext;
//   }

//   private registerJSONSources(jsonSources: any[]): void {
//     if (undefined === jsonSources || null === jsonSources) {
//       return;
//     }
//     let instanceFactory = this._applicationContext.instanceFactory;
//     let jsons = Object.assign({}, ...jsonSources);
//     instanceFactory.registerJSON(jsons);
//   }

//   private registerConfigurations(configurationTypes: ClassType[]): void {
//     if (undefined === configurationTypes || null === configurationTypes) {
//       return;
//     }
//     let instanceFactory = this._applicationContext.instanceFactory;

//     configurationTypes.forEach(configurationType => {
//       let clazz: Class = Class.forClassType(configurationType);
//       let configAnnotation: ConfigurationAnnotation = clazz.getOwnAnnotation(ConfigurationAnnotation);
//       if (undefined === configAnnotation) {
//         throw new Error(`Class is not Configuration type. add @Configuration annotation to class[${configurationType.name}]`);
//       }
//       let methods = clazz.getOwnMethods();
//       if (null === methods) {
//         return;
//       }
//       let instance = clazz.getConstructor().newInstance();
//       instanceFactory.applyInstance(configurationType, instance);
//       methods.forEach(method => {
//         let instanceAnnotation = method.getOwnAnnotation(InstanceAnnotation);
//         if (undefined === instanceAnnotation) {
//           return;
//         }
//         let singletonName = null === instanceAnnotation.attributes.name ? method.getName() : instanceAnnotation.attributes.name;
//         let result = method.invoke(instance);
//         instanceFactory.registerSingleton(singletonName, method.getReturnType(), result);
//       });
//     });
//   }
}
