import {
  Annotation,
  Class,
  Method,
} from '@overflow/commons/core/reflect';

import {
  ClassType,
  PropertyKeyType,
  NotDecoratedClassError,
} from '@overflow/commons/core/type';

import {
  Assert,
  AnnotationUtils,
} from '@overflow/commons/core/util';

import {
  InjectableAnnotation,
  QualifierAnnotation,
} from '@overflow/commons/di/decorators';

import {
  InstanceDefinition,
  InstanceDefinitionRegistry,
  InstanceNameGenerator,
} from '@overflow/commons/di/factory';

import {
  InstanceNameType,
  NotInjectableError,
} from '@overflow/commons/di/type';

import {
  ConfigurationAnnotation,
  InstanceAnnotation,
} from '@overflow/commons/application/decorators';

import {
  NotConfigurationError,
} from '@overflow/commons/application/type';

import { ScopeMetadataResolver } from './scope_metadata_resolver';

export class InstanceDefinitionReader {
  private instanceDefinitionRegistry: InstanceDefinitionRegistry;
  private instanceNameGenerator: InstanceNameGenerator = new InstanceNameGenerator();
  private scopeMetadataResolver: ScopeMetadataResolver = new ScopeMetadataResolver();

  public constructor(registry: InstanceDefinitionRegistry) {
    Assert.notNull(registry, 'InstanceDefinitionRegistry must not be null');
    this.instanceDefinitionRegistry = registry;
  }

  public registerInjectables(...injectables: ClassType[]): void {
    injectables.forEach(injectable => {
      this.registerInjectable(injectable, undefined);
    });
  }

  public registerInjectable(injectable: ClassType, name: InstanceNameType | undefined): InstanceNameType {
    let clazz = Class.forClassType(injectable);
    if (undefined === clazz) {
      throw new NotDecoratedClassError();
    }

    if (!AnnotationUtils.hasAnnotation(injectable, InjectableAnnotation)) {
      throw new NotInjectableError(`Class[${injectable.name}] does not contains @Injectable`);
    }

    let definition: InstanceDefinition = new InstanceDefinition();
    definition.instanceClass = clazz;
    let scopeMeta = this.scopeMetadataResolver.resolveScopeMetadata(definition);
    definition.scope = scopeMeta.scope;
    let instanceName =
                      undefined !== name
                      ? name
                      : this.instanceNameGenerator.generateInstanceName(definition, this.instanceDefinitionRegistry);

    this.instanceDefinitionRegistry.registerInstanceDefinition(instanceName, definition);

    return instanceName;
  }

  public registerConfigurations(...configurations: ClassType[]): void {
    configurations.forEach(configuration => {
      this.registerConfiguration(configuration);
    });
  }

  public registerConfiguration(configuration: ClassType): void {
    let clazz = Class.forClassType(configuration);
    if (undefined === clazz) {
      throw new NotDecoratedClassError();
    }

    if (!AnnotationUtils.hasAnnotation(configuration, ConfigurationAnnotation)) {
      throw new NotConfigurationError(`Class[${configuration.name}] does not contains @Configuration`);
    }

    let factoryInstanceName = this.registerInjectable(configuration, undefined);
    let methods = clazz.getMethods();

    methods.forEach((method: Method, methodName: PropertyKeyType) => {
      let ia = method.getAnnotation(InstanceAnnotation);
      if (undefined === ia) {
        return;
      }
      let instanceName = undefined !== ia.attributes.name ? ia.attributes.name : method.getName();

      let definition: InstanceDefinition = new InstanceDefinition();
      definition.factoryInstanceName = factoryInstanceName;
      definition.factoryMethod = method;

      this.instanceDefinitionRegistry.registerInstanceDefinition(instanceName, definition);
    });
  }
}
